from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0014_no_partially_validated_receiptvalidations'),
    ]

    operations = [
        migrations.RenameField(
            model_name='receiptentry',
            old_name='amount',
            new_name='quantity',
        ),
    ]



"""Tests for AFIP-WS related classes."""
import os
from datetime import date, datetime, timedelta

from django.conf import settings
from django.contrib.auth.models import User
from django.core import management
from django.core.files import File
from django.core.urlresolvers import reverse
from django.test import Client, TestCase
from django.utils.timezone import now

from django_afip import exceptions, models
from testapp.testmain import mocks


class AfipTestCase(TestCase):
    """
    Base class for AFIP-WS related tests.

    Since AFIP rate-limits how often authentication tokens can be fetched, we
    need to keep one between tests.
    This class is a simple hack to keep that ticket in-memory and saves it into
    the DB every time a new class is ``setUp``.
    """

    taxpayer = None
    ticket = None

    def setUp(self):
        """Save a TaxPayer and Ticket into the database."""
        if not AfipTestCase.taxpayer:
            taxpayer = models.TaxPayer(
                pk=1,
                name='test taxpayer',
                cuit=20329642330,
                is_sandboxed=True,
            )

            basepath = settings.BASE_DIR
            with open(os.path.join(basepath, 'test.key')) as key:
                taxpayer.key.save('test.key', File(key))
            with open(os.path.join(basepath, 'test.crt')) as crt:
                taxpayer.certificate.save('test.crt', File(crt))

            AfipTestCase.taxpayer = taxpayer

        if not AfipTestCase.ticket:
            ticket = models.AuthTicket.objects.get_any_active('wsfe')
            AfipTestCase.ticket = ticket

        AfipTestCase.taxpayer.save()
        AfipTestCase.ticket.save()


class AuthTicketTest(TestCase):
    """Test AuthTicket methods."""

    def test_bad_cuit(self):
        """Test using the wrong cuit for a key pair."""

        taxpayer = models.TaxPayer(
            pk=1,
            name='test taxpayer',
            # This is the wrong CUIT for our keypair:
            cuit=20329642339,
            is_sandboxed=True,
        )
        basepath = settings.BASE_DIR
        with open(os.path.join(basepath, 'test.key')) as key:
            taxpayer.key.save('test.key', File(key))
        with open(os.path.join(basepath, 'test.crt')) as crt:
            taxpayer.certificate.save('test.crt', File(crt))
        taxpayer.save()

        taxpayer.create_ticket('wsfe')

        with self.assertRaisesRegex(
            exceptions.AfipException,
            # Note: AFIP apparently edited this message and added a typo:
            'ValidacionDeToken: No apareci[oó] CUIT en lista de relaciones:',
        ):
            models.populate_all()

    def test_bogus_certificate_exception(self):
        """Test that using a junk ceritificates raises as expected."""
        taxpayer = models.TaxPayer(
            pk=1,
            name='test taxpayer',
            cuit=20329642330,
            is_sandboxed=True,
        )
        # Note that we swap key and crt so that it's bogus input:
        basepath = settings.BASE_DIR
        with open(os.path.join(basepath, 'test.crt')) as key:
            taxpayer.key.save('test.key', File(key))
        with open(os.path.join(basepath, 'test.key')) as crt:
            taxpayer.certificate.save('test.crt', File(crt))
        taxpayer.save()

        with self.assertRaises(exceptions.CorruptCertificate) as e:
            taxpayer.create_ticket('wsfe')

        self.assertNotIsInstance(e, exceptions.AfipException)

    def test_no_active_taxpayer(self):
        """Test that no TaxPayers raises an understandable error."""
        with self.assertRaisesMessage(
            exceptions.AuthenticationError,
            'There are no taxpayers to generate a ticket.',
        ):
            models.AuthTicket.objects.get_any_active('wsfe')

    def test_expired_certificate_exception(self):
        """Test that using an expired ceritificate raises as expected."""
        taxpayer = models.TaxPayer(
            pk=1,
            name='test taxpayer',
            cuit=20329642330,
            is_sandboxed=True,
        )
        basepath = settings.BASE_DIR
        with open(os.path.join(basepath, 'test_expired.key')) as key:
            taxpayer.key.save('test.key', File(key))
        with open(os.path.join(basepath, 'test_expired.crt')) as crt:
            taxpayer.certificate.save('test.crt', File(crt))
        taxpayer.save()

        with self.assertRaises(exceptions.CertificateExpired):
            taxpayer.create_ticket('wsfe')

    def test_untrusted_certificate_exception(self):
        """
        Test that using an untrusted ceritificate raises as expected.
        """
        taxpayer = models.TaxPayer(
            pk=1,
            name='test taxpayer',
            cuit=20329642330,
            # Note that we hit production with a sandbox cert here:
            is_sandboxed=False,
        )
        basepath = settings.BASE_DIR
        with open(os.path.join(basepath, 'test.key')) as key:
            taxpayer.key.save('test.key', File(key))
        with open(os.path.join(basepath, 'test.crt')) as crt:
            taxpayer.certificate.save('test.crt', File(crt))
        taxpayer.save()

        with self.assertRaises(exceptions.UntrustedCertificate):
            taxpayer.create_ticket('wsfe')


class PopulationTest(AfipTestCase):
    """
    Tests models population view.

    As a side effect, also test valid ticket generation.
    """

    def test_population_command(self):
        """Test the afipmetadata command."""
        management.call_command('afipmetadata')

        receipts = models.ReceiptType.objects.count()
        concepts = models.ConceptType.objects.count()
        documents = models.DocumentType.objects.count()
        vat = models.VatType.objects.count()
        tax = models.TaxType.objects.count()
        currencies = models.CurrencyType.objects.count()

        self.assertGreater(receipts, 0)
        self.assertGreater(concepts, 0)
        self.assertGreater(documents, 0)
        self.assertGreater(vat, 0)
        self.assertGreater(tax, 0)
        self.assertGreater(currencies, 0)


class TaxPayerTest(AfipTestCase):
    """Test TaxPayer methods."""

    def test_fetch_points_of_sale(self):
        """Test the ``fetch_points_of_sales`` method."""
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()

        points_of_sales = models.PointOfSales.objects.count()
        self.assertGreater(points_of_sales, 0)


class ReceiptBatchTest(AfipTestCase):
    """Test ReceiptBatch methods."""

    def setUp(self):
        """Populate AFIP metadata and create a TaxPayer and PointOfSales."""
        super().setUp()
        models.populate_all()
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()

    def _good_receipt(self):
        return mocks.receipt()

    def _bad_receipt(self):
        return mocks.receipt(80)

    def test_creation_empty(self):
        """
        Test creation of an empty batch.

        An empty batch has no sense, and None should be returned.
        """
        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.none(),
        )
        self.assertIsNone(batch)

    def test_creation_exclusion(self):
        """Test that batch creation excludes already batched receipts."""
        self._good_receipt()
        self._good_receipt()
        self._good_receipt()

        models.ReceiptBatch.objects.create(
            models.Receipt.objects.filter(pk=1),
        )
        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.filter(batch_id__isnull=True),
        )

        self.assertEquals(batch.receipts.count(), 2)

    def test_validate_empty(self):
        """Test that validating an empty batch does not crash."""
        # Hack to easily create an empty batch:
        self._good_receipt()
        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.all(),
        )
        models.Receipt.objects.all().delete()

        errs = batch.validate()
        self.assertEqual(errs, [])

    def test_validation_good(self):
        """Test validating valid receipts."""
        self._good_receipt()
        self._good_receipt()
        self._good_receipt()

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            batch.validation.last().result,
            models.Validation.RESULT_APPROVED,
        )
        self.assertEqual(batch.validation.count(), 1)
        self.assertEqual(batch.receipts.count(), 3)

    def test_validation_bad(self):
        """Test validating invalid receipts."""
        self._bad_receipt()
        self._bad_receipt()
        self._bad_receipt()

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            'Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: '
            '80, DocNro 203012345 no se encuentra registrado en los padrones '
            'de AFIP y no corresponde a una cuit pais.'
        )
        self.assertEqual(batch.receipts.count(), 0)

    def test_validation_mixed(self):
        """
        Test validating a mixture of valid and invalid receipts.

        Receipts are validated by AFIP in-order, so all receipts previous to
        the bad one are validated, and nothing else is even parsed after the
        invalid one.
        """
        self._good_receipt()
        self._bad_receipt()
        self._good_receipt()

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            'Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: '
            '80, DocNro 203012345 no se encuentra registrado en los padrones '
            'de AFIP y no corresponde a una cuit pais.'
        )
        self.assertEqual(batch.receipts.count(), 1)

    def test_validation_validated(self):
        """Test validating invalid receipts."""
        receipt = self._good_receipt()
        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.all()
        )
        validation = models.Validation.objects.create(
            processed_date=now(),
            result=models.Validation.RESULT_APPROVED,
            batch=batch,
        )
        models.ReceiptValidation.objects.create(
            validation=validation,
            result=models.Validation.RESULT_APPROVED,
            cae='123',
            cae_expiration=now(),
            receipt=receipt,
        )

        receipt.batch = None
        receipt.save()

        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.all()
        )
        self.assertIsNotNone(batch)
        errs = batch.validate()

        self.assertEqual(errs, [])

    def test_validation_good_service(self):
        """Test validating a receipt for a service (rather than product)."""
        receipt = self._good_receipt()
        receipt.concept_id = 2
        receipt.service_start = datetime.now() - timedelta(days=10)
        receipt.service_end = datetime.now()
        receipt.expiration_date = datetime.now() + timedelta(days=10)
        receipt.save()

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            batch.validation.last().result,
            models.Validation.RESULT_APPROVED,
        )
        self.assertEqual(batch.receipts.count(), 1)

    def test_validation_good_without_tax(self):
        """Test validating valid receipts."""
        mocks.receipt(with_tax=False)

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            batch.validation.last().result,
            models.Validation.RESULT_APPROVED,
        )
        self.assertEqual(batch.validation.count(), 1)
        self.assertEqual(batch.receipts.count(), 1)

    def test_validation_good_without_vat(self):
        """Test validating valid receipts."""
        mocks.receipt(with_vat=False, receipt_type=11)

        batch = models.ReceiptBatch.objects \
            .create(models.Receipt.objects.all())
        errs = batch.validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            batch.validation.last().result,
            models.Validation.RESULT_APPROVED,
        )
        self.assertEqual(batch.validation.count(), 1)
        self.assertEqual(batch.receipts.count(), 1)


class ReceiptPDFTest(AfipTestCase):
    """Test ReceiptPDF methods."""

    # TODO: Test generation via ReceiptHTMLView

    def setUp(self):
        """Generate a valid receipt for later PDF generation."""
        super().setUp()
        models.populate_all()
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()
        models.TaxPayerProfile.objects.create(
            taxpayer=taxpayer,
            issuing_name='Red Company Inc.',
            issuing_address='100 Red Av\nRedsville\nUK',
            issuing_email='billing@example.com',
            vat_condition='Exempt',
            gross_income_condition='Exempt',
            sales_terms='Credit Card',
            active_since=datetime(2011, 10, 3),
        )

        models.Receipt.objects.create(
            concept=models.ConceptType.objects.get(code=1),
            document_type=models.DocumentType.objects.get(code=96),
            document_number='203012345',
            issued_date=date.today(),
            total_amount=100,
            net_untaxed=0,
            net_taxed=100,
            exempt_amount=0,
            currency=models.CurrencyType.objects.get(code='PES'),
            currency_quote=1,

            receipt_number=4236,
            receipt_type=models.ReceiptType.objects.get(code=11),
            point_of_sales=models.PointOfSales.objects.first(),
        )

        # TODO: Add a ReceiptEntry

    def _create_valid_receipt(self):
        receipt = models.Receipt.objects.first()
        pdf = models.ReceiptPDF.objects.create_for_receipt(
            receipt=receipt,
            client_name='John Doe',
            client_address='12 Green Road\nGreenville\nUK',
        )
        pdf.save_pdf()

        return pdf

    def test_pdf_generation(self):
        """
        Test PDF file generation.

        For the moment, this test case mostly verifies that pdf generation
        *works*, but does not actually validate the pdf file itself.

        Running this locally *will* yield the file itself, which is useful for
        manual inspection.
        """
        pdf = self._create_valid_receipt()
        self.assertTrue(pdf.pdf_file.name.startswith('receipts/'))
        self.assertTrue(pdf.pdf_file.name.endswith('.pdf'))

    def test_html_view(self):
        """
        Test the PDF generation view.
        """
        pdf = self._create_valid_receipt()

        client = Client()
        response = client.get(
            reverse('receipt_html_view', args=(pdf.receipt.pk,))
        )
        self.assertContains(
            response,
            '<div class="client">\n<strong>Facturado a:</strong><br>\n John '
            'Doe,\nDNI\n203012345\n<!-- linkbreaks adds a br here -->\n<p>12 '
            'Green Road<br />Greenville<br />UK</p>\n Exempt<br>\nCredit '
            'Card\n</div>',
            html=True,
        )

    def test_pdf_view(self):
        """
        Test the PDF generation view.
        """
        pdf = self._create_valid_receipt()

        client = Client()
        response = client.get(
            reverse('receipt_pdf_view', args=(pdf.receipt.pk,))
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content[:10], b'%PDF-1.5\n%')

        headers = sorted(response.serialize_headers().decode().splitlines())
        expected_headers = sorted([
            'Content-Type: application/pdf',
            'X-Frame-Options: SAMEORIGIN',
            'Content-Disposition: attachment; filename=receipt {}.pdf'.format(
                pdf.receipt.pk
            ),
        ])
        self.assertEqual(headers, expected_headers)

    def test_unauthorized_receipt_generation(self):
        """
        Test PDF file generation for unauthorized receipts.

        Confirm that attempting to generate a PDF for an unauthorized receipt
        raises.
        """
        receipt = models.Receipt.objects.first()
        receipt.receipt_number = None
        receipt.save()
        pdf = models.ReceiptPDF.objects.create_for_receipt(
            receipt=receipt,
            client_name='John Doe',
            client_address='12 Green Road\nGreenville\nUK',
        )
        with self.assertRaisesMessage(
            Exception,
            'Cannot generate pdf for non-authorized receipt'
        ):
            pdf.save_pdf()


class ReceiptAdminTest(AfipTestCase):
    """Test ReceiptAdmin methods."""

    def setUp(self):
        """Initialized AFIP metadata and a single django superuser."""
        super().setUp()
        models.populate_all()
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()

        User.objects._create_user(
           username='superuser',
           email='superuser@email.com',
           password='123',
           is_staff=True,
           is_superuser=True,
        )

    def test_validation_filters(self):
        """
        Test the admin validation filters.

        This filters receipts by the validation status.
        """
        validated_receipt = mocks.receipt()
        not_validated_receipt = mocks.receipt()
        # XXX: Receipt with failed validation?

        batch = models.ReceiptBatch.objects.create(
            models.Receipt.objects.filter(pk=validated_receipt.pk)
        )
        validation = models.Validation.objects.create(
            processed_date=now(),
            result=models.Validation.RESULT_APPROVED,
            batch=batch,
        )
        models.ReceiptValidation.objects.create(
            validation=validation,
            result=models.Validation.RESULT_APPROVED,
            cae='123',
            cae_expiration=now(),
            receipt=validated_receipt,
        )

        client = Client()
        client.force_login(User.objects.first())

        response = client.get('/admin/afip/receipt/?status=validated')
        self.assertContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(validated_receipt.pk),
            html=True,
        )
        self.assertNotContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(not_validated_receipt.pk),
            html=True,
        )

        response = client.get('/admin/afip/receipt/?status=not_validated')
        self.assertNotContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(validated_receipt.pk),
            html=True,
        )
        self.assertContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(not_validated_receipt.pk),
            html=True,
        )


# TODO: Test receipts with related_receipts



"""
This file monkey-patches HTTPSConnection so that it works with AFIP's sandbox
servers, which advertises capabilities it does not have.

It attempts to be as compatible as possible, and as little invasible as
possible.

Alternate solutions to the underlying issue are very much welcome.
"""
import http.client
import logging
import socket
import ssl
from http.client import HTTPConnection, HTTPS_PORT
from ssl import _ssl, CERT_REQUIRED, OP_NO_SSLv2, OP_NO_SSLv3, \
    PROTOCOL_TLSv1, Purpose, SSLContext

logger = logging.getLogger(__name__)


def create_afip_compatible_context(purpose=Purpose.SERVER_AUTH, *, cafile=None,
                                   capath=None, cadata=None):
    """Monkeys and dragons be here!"""
    context = SSLContext(PROTOCOL_TLSv1)
    context.options |= OP_NO_SSLv2
    context.options |= OP_NO_SSLv3
    context.options |= getattr(_ssl, 'OP_NO_COMPRESSION', 0)

    context.verify_mode = CERT_REQUIRED
    context.check_hostname = True

    context.load_default_certs(purpose)
    return context


class MonkeyPatchedHTTPSConnection(HTTPConnection):

    default_port = HTTPS_PORT

    # XXX Should key_file and cert_file be deprecated in favour of context?

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 source_address=None, *, context=None,
                 check_hostname=None):
        super().__init__(host, port, timeout, source_address)
        self.key_file = key_file
        self.cert_file = cert_file
        if context is None:
            if host == 'wswhomo.afip.gov.ar':
                context = create_afip_compatible_context()
            elif hasattr(ssl, '_create_default_https_context'):  # py 3.4.3
                context = ssl._create_default_https_context()
            else:  # py 3.4.2
                context = ssl._create_stdlib_context()
        will_verify = context.verify_mode != ssl.CERT_NONE
        if check_hostname is None:
            check_hostname = context.check_hostname
        if check_hostname and not will_verify:
            raise ValueError('check_hostname needs a SSL context with '
                             'either CERT_OPTIONAL or CERT_REQUIRED')
        if key_file or cert_file:
            context.load_cert_chain(cert_file, key_file)
        self._context = context
        self._check_hostname = check_hostname

    def connect(self):
        """Connect to a host on a given (SSL) port."""

        super().connect()

        if self._tunnel_host:
            server_hostname = self._tunnel_host
        else:
            server_hostname = self.host

        self.sock = self._context.wrap_socket(self.sock,
                                              server_hostname=server_hostname)
        if not self._context.check_hostname and self._check_hostname:
            try:
                ssl.match_hostname(self.sock.getpeercert(), server_hostname)
            except Exception:
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
                raise


def patch_https_for_afip():
    logger.info(
        "Monkey patching HTTPSConnection to work with AFIP's servers"
    )
    http.client.HTTPSConnection = MonkeyPatchedHTTPSConnection



from unittest import mock

from django.contrib import messages
from django.http import HttpRequest
from django.test import Client, TestCase
from django.utils.translation import ugettext as _

from django_afip import exceptions
from django_afip.admin import catch_errors
from testapp.testmain import mocks


class TestCatchErrors(TestCase):

    def _get_test_instance(self, exception_type):

        class TestClass(mock.MagicMock):
            @catch_errors
            def action(self, request):
                raise exception_type

        return TestClass()

    def test_certificate_expired(self):
        obj = self._get_test_instance(exceptions.CertificateExpired)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('The AFIP Taxpayer certificate has expired.'),
            messages.ERROR,
        ))

    def test_certificate_untrusted_cert(self):
        obj = self._get_test_instance(exceptions.UntrustedCertificate)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('The AFIP Taxpayer certificate is untrusted.'),
            messages.ERROR,
        ))

    def test_certificate_auth_error(self):
        obj = self._get_test_instance(exceptions.AuthenticationError)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('An unknown authentication error has ocurred: '),
            messages.ERROR,
        ))


class TestTaxPayerAdminKeyGeneration(TestCase):

    def setUp(self):
        self.user = mocks.superuser()

    def test_without_key(self):
        taxpayer = mocks.taxpayer()
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data=dict(
            _selected_action=[taxpayer.id],
            action='generate_key',
        ), follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Key generated successfully.')

        taxpayer.refresh_from_db()
        self.assertIn(
            '-----BEGIN PRIVATE KEY-----',
            taxpayer.key.file.read().decode(),
        )

    def test_with_key(self):
        taxpayer = mocks.taxpayer(key='Blah'.encode())
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data=dict(
            _selected_action=[taxpayer.id],
            action='generate_key',
        ), follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No keys generated; Taxpayers already had keys.',
        )

        taxpayer.refresh_from_db()
        self.assertEqual('Blah', taxpayer.key.file.read().decode())


class TestTaxPayerAdminRequestGeneration(TestCase):

    def setUp(self):
        self.user = mocks.superuser()

    def test_with_csr(self):
        taxpayer = mocks.taxpayer()
        taxpayer.generate_key()
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data=dict(
            _selected_action=[taxpayer.id],
            action='generate_csr',
        ), follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            b'Content-Type: application/pkcs10',
            response.serialize_headers().splitlines()
        )
        self.assertContains(
            response,
            '-----BEGIN CERTIFICATE REQUEST-----',
        )

    def test_without_key(self):
        taxpayer = mocks.taxpayer()
        taxpayer.generate_key()
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data=dict(
            _selected_action=[taxpayer.id],
            action='generate_csr',
        ), follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            b'Content-Type: application/pkcs10',
            response.serialize_headers().splitlines()
        )
        self.assertContains(
            response,
            '-----BEGIN CERTIFICATE REQUEST-----',
        )

    def test_multiple_taxpayers(self):
        taxpayer1 = mocks.taxpayer(key='Blah'.encode())
        taxpayer2 = mocks.taxpayer(key='Blah'.encode())
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data=dict(
            _selected_action=[taxpayer1.id, taxpayer2.id],
            action='generate_csr',
        ), follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'Can only generate CSR for one taxpayer at a time',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-01-04 21:13
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0009_auto_20151214_1836'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receiptpdf',
            name='pdf_file',
            field=models.FileField(blank=True, null=True, upload_to='receipts', verbose_name='pdf file'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-12 14:31
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_afip.models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0005_auto_20150918_0115'),
    ]

    operations = [
        migrations.CreateModel(
            name='ReceiptEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=128, verbose_name='description')),
                ('amount', models.PositiveSmallIntegerField(verbose_name='amount')),
                ('unit_price', models.DecimalField(decimal_places=2, max_digits=15, verbose_name='unit price')),
            ],
            options={
                'verbose_name': 'receipt entry',
                'verbose_name_plural': 'receipt entries',
            },
        ),
        migrations.CreateModel(
            name='ReceiptPDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pdf_file', models.FileField(null=True, upload_to='', verbose_name='pdf file')),
                ('issuing_name', models.TextField(verbose_name='issuing name')),
                ('issuing_address', models.TextField(verbose_name='issuing address')),
                ('issuing_email', models.TextField(null=True, verbose_name='issuing address')),
                ('vat_condition', models.CharField(max_length=48, verbose_name='vat condition')),
                ('gross_income_condition', models.CharField(max_length=48, verbose_name='gross income condition')),
                ('client_name', models.TextField(verbose_name='client name')),
                ('client_address', models.TextField(verbose_name='client address')),
                ('sales_terms', models.CharField(max_length=48, verbose_name='sales terms')),
            ],
            options={
                'verbose_name': 'receipt pdf',
                'verbose_name_plural': 'receipt pdfs',
            },
        ),
        migrations.CreateModel(
            name='TaxPayerProfile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('issuing_name', models.TextField(verbose_name='issuing name')),
                ('issuing_address', models.TextField(verbose_name='issuing address')),
                ('vat_condition', models.CharField(max_length=48, verbose_name='vat condition')),
                ('gross_income_condition', models.CharField(max_length=48, verbose_name='gross income condition')),
            ],
        ),
        migrations.AlterModelOptions(
            name='authticket',
            options={'verbose_name': 'authorization ticket', 'verbose_name_plural': 'authorization tickets'},
        ),
        migrations.AlterModelOptions(
            name='concepttype',
            options={'verbose_name': 'concept type', 'verbose_name_plural': 'concept types'},
        ),
        migrations.AlterModelOptions(
            name='currencytype',
            options={'verbose_name': 'currency type', 'verbose_name_plural': 'currency types'},
        ),
        migrations.AlterModelOptions(
            name='documenttype',
            options={'verbose_name': 'document type', 'verbose_name_plural': 'document types'},
        ),
        migrations.AlterModelOptions(
            name='observation',
            options={'verbose_name': 'observation', 'verbose_name_plural': 'observations'},
        ),
        migrations.AlterModelOptions(
            name='pointofsales',
            options={'verbose_name': 'point of sales', 'verbose_name_plural': 'points of sales'},
        ),
        migrations.AlterModelOptions(
            name='receipt',
            options={'verbose_name': 'receipt', 'verbose_name_plural': 'receipts'},
        ),
        migrations.AlterModelOptions(
            name='receiptbatch',
            options={'verbose_name': 'receipt batch', 'verbose_name_plural': 'receipt batches'},
        ),
        migrations.AlterModelOptions(
            name='receipttype',
            options={'verbose_name': 'receipt type', 'verbose_name_plural': 'receipt types'},
        ),
        migrations.AlterModelOptions(
            name='receiptvalidation',
            options={'verbose_name': 'receipt validation', 'verbose_name_plural': 'receipt validations'},
        ),
        migrations.AlterModelOptions(
            name='tax',
            options={'verbose_name': 'tax', 'verbose_name_plural': 'taxes'},
        ),
        migrations.AlterModelOptions(
            name='taxpayer',
            options={'verbose_name': 'taxpayer', 'verbose_name_plural': 'taxpayers'},
        ),
        migrations.AlterModelOptions(
            name='taxtype',
            options={'verbose_name': 'tax type', 'verbose_name_plural': 'tax types'},
        ),
        migrations.AlterModelOptions(
            name='validation',
            options={'verbose_name': 'validation', 'verbose_name_plural': 'validations'},
        ),
        migrations.AlterModelOptions(
            name='vat',
            options={'verbose_name': 'vat', 'verbose_name_plural': 'vat'},
        ),
        migrations.AlterModelOptions(
            name='vattype',
            options={'verbose_name': 'vat type', 'verbose_name_plural': 'vat types'},
        ),
        migrations.AddField(
            model_name='taxpayer',
            name='active_since',
            field=models.DateField(help_text='Date since which this taxpayer has been legally active. This field is only required to generate receipt PDFs', null=True, verbose_name='active since'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='expires',
            field=models.DateTimeField(default=django_afip.models.AuthTicket.default_expires, verbose_name='expires'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='generated',
            field=models.DateTimeField(default=django_afip.models.AuthTicket.default_generated, verbose_name='generated'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tickets', to='afip.TaxPayer', verbose_name='owner'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='service',
            field=models.CharField(help_text='Service for which this ticket has been authorized', max_length=6, verbose_name='service'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='signature',
            field=models.TextField(verbose_name='signature'),
        ),
        migrations.AlterField(
            model_name='authticket',
            name='unique_id',
            field=models.IntegerField(default=django_afip.models.AuthTicket.default_unique_id, verbose_name='unique id'),
        ),
        migrations.AlterField(
            model_name='concepttype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='concepttype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='concepttype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='concepttype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AlterField(
            model_name='currencytype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='currencytype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='currencytype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='currencytype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AlterField(
            model_name='observation',
            name='code',
            field=models.PositiveSmallIntegerField(verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='observation',
            name='message',
            field=models.CharField(max_length=255, verbose_name='message'),
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='blocked',
            field=models.BooleanField(verbose_name='blocked'),
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='drop_date',
            field=models.DateField(blank=True, null=True, verbose_name='drop date'),
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='issuance_type',
            field=models.CharField(help_text='Indicates if thie POS emits using CAE and CAEA.', max_length=8, verbose_name='issuance type'),
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='number',
            field=models.PositiveSmallIntegerField(verbose_name='number'),
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.TaxPayer', verbose_name='owner'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='batch',
            field=models.ForeignKey(blank=True, help_text='Receipts are validated in batches, so it must be assigned one before validation is possible.', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.ReceiptBatch', verbose_name='receipt batch'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='concept',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.ConceptType', verbose_name='concept'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='currency',
            field=models.ForeignKey(help_text='Currency in which this receipt is issued.', on_delete=django.db.models.deletion.CASCADE, related_name='documents', to='afip.CurrencyType', verbose_name='currency'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='currency_quote',
            field=models.DecimalField(decimal_places=6, help_text='Quote of the day for the currency used in the receipt', max_digits=10, verbose_name='currency quote'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='document_number',
            field=models.BigIntegerField(help_text='The document number of the customer to whom this receipt is addressed', verbose_name='document number'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='document_type',
            field=models.ForeignKey(help_text='The document type of the customer to whom this receipt is addressed', on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.DocumentType', verbose_name='document type'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='exempt_amount',
            field=models.DecimalField(decimal_places=2, help_text='Only for categories which are tax-exempt. For C-type receipts, this must be zero.', max_digits=15, verbose_name='exempt amount'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='expiration_date',
            field=models.DateField(blank=True, help_text='Date on which this receipt expires. No applicable for goods.', null=True, verbose_name='receipt expiration date'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='issued_date',
            field=models.DateField(help_text='Can diverge up to 5 days for good, or 10 days otherwise', verbose_name='issued date'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='net_taxed',
            field=models.DecimalField(decimal_places=2, help_text='The total amount to which taxes apply. For C-type receipts, this is equal to the subtotal.', max_digits=15, verbose_name='total taxable amount'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='net_untaxed',
            field=models.DecimalField(decimal_places=2, help_text='The total amount to which taxes do not apply. For C-type receipts, this must be zero.', max_digits=15, verbose_name='total untaxable amount'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='point_of_sales',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.PointOfSales', verbose_name='point of sales'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='receipt_number',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='receipt number'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='receipt_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.ReceiptType', verbose_name='receipt type'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='related_receipts',
            field=models.ManyToManyField(blank=True, to='afip.Receipt', verbose_name='related receipts'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='service_end',
            field=models.DateField(blank=True, help_text='Date on which a service ended. No applicable for goods.', null=True, verbose_name='service end date'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='service_start',
            field=models.DateField(blank=True, help_text='Date on which a service started. No applicable for goods.', null=True, verbose_name='service start date'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='total_amount',
            field=models.DecimalField(decimal_places=2, help_text='Must be equal to untaxed amount + exempt amount + taxes + vat.', max_digits=15, verbose_name='total amount'),
        ),
        migrations.AlterField(
            model_name='receiptbatch',
            name='point_of_sales',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.PointOfSales', verbose_name='point of sales'),
        ),
        migrations.AlterField(
            model_name='receiptbatch',
            name='receipt_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.ReceiptType', verbose_name='receipt type'),
        ),
        migrations.AlterField(
            model_name='receipttype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='receipttype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='receipttype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='receipttype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='cae_expiration',
            field=models.DateField(verbose_name='cae expiration'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='observations',
            field=models.ManyToManyField(related_name='valiations', to='afip.Observation', verbose_name='observations'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='receipt',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='validation', to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='result',
            field=models.CharField(choices=[('A', 'approved'), ('R', 'rejected'), ('P', 'partial')], max_length=1, verbose_name='result'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='validation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.Validation', verbose_name='validation'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='aliquot',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='aliquot'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='amount'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='base_amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='base amount'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='description',
            field=models.CharField(max_length=80, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='tax_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.TaxType', verbose_name='tax type'),
        ),
        migrations.AlterField(
            model_name='taxpayer',
            name='certificate',
            field=models.FileField(null=True, upload_to='', verbose_name='certificate'),
        ),
        migrations.AlterField(
            model_name='taxpayer',
            name='key',
            field=models.FileField(null=True, upload_to='', verbose_name='key'),
        ),
        migrations.AlterField(
            model_name='taxpayer',
            name='name',
            field=models.CharField(help_text='A friendly name to recognize this taxpayer.', max_length=32, verbose_name='name'),
        ),
        migrations.AlterField(
            model_name='taxtype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='taxtype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='taxtype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='taxtype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AlterField(
            model_name='validation',
            name='batch',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='validation', to='afip.ReceiptBatch', verbose_name='receipt batch'),
        ),
        migrations.AlterField(
            model_name='validation',
            name='processed_date',
            field=models.DateTimeField(verbose_name='processed date'),
        ),
        migrations.AlterField(
            model_name='validation',
            name='result',
            field=models.CharField(choices=[('A', 'approved'), ('R', 'rejected'), ('P', 'partial')], max_length=1, verbose_name='result'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='amount'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='base_amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='base amount'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='vat_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.VatType', verbose_name='vat type'),
        ),
        migrations.AlterField(
            model_name='vattype',
            name='code',
            field=models.CharField(max_length=3, verbose_name='code'),
        ),
        migrations.AlterField(
            model_name='vattype',
            name='description',
            field=models.CharField(max_length=250, verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='vattype',
            name='valid_from',
            field=models.DateField(blank=True, null=True, verbose_name='valid from'),
        ),
        migrations.AlterField(
            model_name='vattype',
            name='valid_to',
            field=models.DateField(blank=True, null=True, verbose_name='valid until'),
        ),
        migrations.AddField(
            model_name='taxpayerprofile',
            name='taxpayer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to='afip.TaxPayer', verbose_name='taxpayer'),
        ),
        migrations.AddField(
            model_name='receiptpdf',
            name='receipt',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AddField(
            model_name='receiptentry',
            name='receipt',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='afip.Receipt', verbose_name='receipt'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-02-03 02:56
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0010_receiptpdf_pdffile_uploadto'),
    ]

    operations = [
        migrations.AddField(
            model_name='receiptentry',
            name='vat',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='receipt_entries', to='afip.VatType', verbose_name='vat'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-06 07:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0012_taxpayer_profile_one_to_one'),
    ]

    operations = [
        migrations.AddField(
            model_name='taxpayer',
            name='is_sandboxed',
            field=models.BooleanField(default=True, help_text='Indicates if this taxpayer interacts with the sandbox servers rather than the production servers', verbose_name='is sandboxed'),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-10 13:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0016_auto_20170529_2012'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receipt',
            name='issued_date',
            field=models.DateField(
                help_text=(
                    'Can diverge up to 5 days for good, or 10 days otherwise'
                ),
                verbose_name='issued date',
            ),
        ),
    ]



import functools
import logging
from datetime import datetime

from django.apps import apps
from django.contrib import admin, messages
from django.contrib.admin.sites import AlreadyRegistered
from django.core.urlresolvers import reverse
from django.db.models import Count
from django.http import HttpResponse
from django.utils.translation import ugettext as _

from django_afip import exceptions, models


logger = logging.getLogger(__name__)


# TODO: Add an action to populate generic types.


def catch_errors(f):
    """
    Catches specific errors in admin actions and shows a friendly error.
    """

    @functools.wraps(f)
    def wrapper(self, request, *args, **kwargs):
        try:
            return f(self, request, *args, **kwargs)
        except exceptions.CertificateExpired as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate has expired.'),
                messages.ERROR,
            )
        except exceptions.UntrustedCertificate as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate is untrusted.'),
                messages.ERROR,
            )
        except exceptions.CorruptCertificate as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate is corrupt.'),
                messages.ERROR,
            )
        except exceptions.AuthenticationError as e:
            logger.exception('AFIP auth failed')
            self.message_user(
                request,
                _('An unknown authentication error has ocurred: %s') % e,
                messages.ERROR,
            )

    return wrapper


class VatInline(admin.TabularInline):
    model = models.Vat
    fields = (
        'vat_type',
        'base_amount',
        'amount',
    )
    extra = 1


class TaxInline(admin.TabularInline):
    model = models.Tax
    fields = (
        'tax_type',
        'description',
        'base_amount',
        'aliquot',
        'amount',
    )
    extra = 1


class ReceiptStatusFilter(admin.SimpleListFilter):
    title = _('status')
    parameter_name = 'status'

    VALIDATED = 'validated'
    NOT_VALIDATED = 'not_validated'

    def lookups(self, request, model_admin):
        return (
            (self.VALIDATED, _('Validated')),
            (self.NOT_VALIDATED, _('Not validated')),
        )

    def queryset(self, request, queryset):
        if self.value() == self.VALIDATED:
            return queryset.filter(
                validation__result=models.Validation.RESULT_APPROVED
            )
        if self.value() == self.NOT_VALIDATED:
            return queryset.exclude(
                validation__result=models.Validation.RESULT_APPROVED
            )


class ReceiptBatchedFilter(admin.SimpleListFilter):
    title = _('batched')
    parameter_name = 'batched'

    BATCHED = 'batched'
    NOT_BATCHED = 'not_batched'

    def lookups(self, request, model_admin):
        return (
            (self.BATCHED, _('Batched')),
            (self.NOT_BATCHED, _('Not batched')),
        )

    def queryset(self, request, queryset):
        if self.value() == self.BATCHED:
            return queryset.filter(
                batch__isnull=True,
            )
        if self.value() == self.NOT_BATCHED:
            return queryset.filter(
                batch__isnull=False
            )


class ReceiptAdmin(admin.ModelAdmin):

    list_display = (
        'id',
        'receipt_type',
        'point_of_sales',
        'receipt_number',
        'issued_date',
        'total_amount',
        'batch_link',
        'validated',
    )
    list_filter = (
        ReceiptBatchedFilter,
        ReceiptStatusFilter,
        'batch',
    )

    __related_fields = [
        'validated',
        'cae',
    ]

    filter_horizontal = ('related_receipts',)

    inlines = (
        VatInline,
        TaxInline,
    )
    ordering = (
        'id',
    )

    def get_fields(self, request, obj=None):
        return super().get_fields(request, obj)

    readonly_fields = __related_fields

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .select_related(
                'validation',
                'receipt_type',
                'point_of_sales',
                'batch',
            )

    def validated(self, obj):
        return obj.validation.result == models.Validation.RESULT_APPROVED
    validated.short_description = _('validated')
    validated.admin_order_field = 'validation__result'
    validated.boolean = True

    def cae(self, obj):
        return obj.validation.cae
    cae.short_description = _('cae')
    cae.admin_order_field = 'validation__cae'

    @catch_errors
    def create_batch(self, request, queryset):
        variations = queryset \
            .aggregate(
                receipt_types=Count('receipt_type', distinct=True),
                points_of_sales=Count('point_of_sales', distinct=True),
            )

        if variations['receipt_types'] > 1:
            self.message_user(
                request,
                _(
                    'The selected receipts are not all of the same type.'
                ),
                messages.ERROR,
            )
            return
        if variations['points_of_sales'] > 1:
            self.message_user(
                request,
                _(
                    'The selected receipts are not all from the same point '
                    'of sales.'
                ),
                messages.ERROR,
            )
            return

        queryset = queryset.filter(batch__isnull=True)
        if not queryset.count():
            self.message_user(
                request,
                _('All the selected invoices have already been batched'),
                messages.ERROR,
            )
            return

        models.ReceiptBatch.objects.create(queryset)
        # TODO: Maybe redirect to batch screen?
    create_batch.short_description = _('Create receipt batch')

    def batch_link(self, obj):
        if not obj.batch:
            return None
        return '<a href="{}">{}</a>'.format(
            reverse('admin:afip_receiptbatch_change', args=(obj.batch.id,)),
            obj.batch.id
        )
    batch_link.admin_order_field = 'batch'
    batch_link.allow_tags = True
    batch_link.short_description = _('batch')

    actions = [create_batch]


class ReceiptBatchAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'receipts_count',
        'validated',
    )

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.annotate(
            Count('receipts', distinct=True)
        )

    def validated(self, obj):
        return obj.validation \
            .filter(result=models.Validation.RESULT_APPROVED) \
            .count() > 0
    validated.short_description = _('validated')
    validated.admin_order_field = 'validation__result'
    validated.boolean = True

    def receipts_count(self, obj):
        return '<a href="{}?batch__id__exact={}">{}</a>'.format(
            reverse(self.admin_site.name + ':afip_receipt_changelist'),
            obj.id,
            obj.receipts__count,
        )
    receipts_count.allow_tags = True
    receipts_count.short_description = _('receipts')
    receipts_count.admin_order_field = 'receipts__count'

    @catch_errors
    def validate(self, request, queryset):
        for batch in queryset:
            errs = batch.validate()
            if errs:
                self.message_user(
                    request,
                    _(
                        'Batch #%(num)s failed: %(err)s'
                    ) % {'num': batch.pk, 'err': errs},
                    messages.ERROR,
                )
    validate.short_description = _('Validate')

    actions = [validate]


class AuthTicketAdmin(admin.ModelAdmin):
    list_display = (
        'unique_id',
        'owner',
        'service',
        'generated',
        'expires',
    )


class TaxPayerAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'cuit',
    )

    @catch_errors
    def fetch_points_of_sales(self, request, queryset):
        poses = [
            pos
            for taxpayer in queryset.all()
            for pos in taxpayer.fetch_points_of_sales()
        ]

        created = len([pos for pos in poses if pos[1]])
        total = len(poses)

        self.message_user(
            request,
            message=(
                _(
                    '%(total)d points of sales fetched. %(created)d created.'
                ) % dict(
                    total=total,
                    created=created,
                )
            ),
            level=messages.SUCCESS,
        )

    fetch_points_of_sales.short_description = _('Fetch points of sales')

    def generate_key(self, request, queryset):
        key_count = sum([t.generate_key() for t in queryset.all()])

        if key_count is 1:
            message = _('Key generated successfully.')
            level = messages.SUCCESS
        elif key_count is 1:
            message = _('%d keys generated successfully.') % key_count
            level = messages.SUCCESS
        else:
            message = _('No keys generated; Taxpayers already had keys.')
            level = messages.ERROR

        self.message_user(
            request,
            message=message,
            level=level,
        )
    generate_key.short_description = _('Generate key')

    def generate_csr(self, request, queryset):
        if queryset.count() > 1:
            self.message_user(
                request,
                message=_('Can only generate CSR for one taxpayer at a time'),
                level=messages.ERROR,
            )
            return

        taxpayer = queryset.first()
        if not taxpayer.key:
            taxpayer.generate_key()

        csr = taxpayer.generate_csr()
        filename = 'cuit-{}-{}.csr'.format(
            taxpayer.cuit,
            int(datetime.now().timestamp()),
        )

        response = HttpResponse(content_type='application/pkcs10')
        response['Content-Disposition'] = 'attachment; filename={}'.format(
            filename
        )

        response.write(csr.read())
        return response
    generate_csr.short_description = _('Generate CSR')

    actions = (
        fetch_points_of_sales,
        generate_key,
        generate_csr,
    )


@admin.register(models.PointOfSales)
class PointOfSalesAdmin(admin.ModelAdmin):
    list_display = (
        'owner',
        'number',
        'issuance_type',
        'blocked',
        'drop_date',
    )


@admin.register(models.CurrencyType)
class CurrencyTypeAdmin(admin.ModelAdmin):
    list_display = (
        'code',
        'description',
        'valid_from',
        'valid_to',
    )


admin.site.register(models.Receipt, ReceiptAdmin)
admin.site.register(models.ReceiptBatch, ReceiptBatchAdmin)
admin.site.register(models.AuthTicket, AuthTicketAdmin)
admin.site.register(models.TaxPayer, TaxPayerAdmin)

app = apps.get_app_config('afip')
for model in app.get_models():
    try:
        admin.site.register(model)
    except AlreadyRegistered:
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-12 17:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0006_auto_20151212_1431'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='taxpayerprofile',
            options={'verbose_name': 'taxpayer profile', 'verbose_name_plural': 'taxpayer profiles'},
        ),
        migrations.RemoveField(
            model_name='taxpayer',
            name='active_since',
        ),
        migrations.AddField(
            model_name='taxpayerprofile',
            name='active_since',
            field=models.DateField(default=None, help_text='Date since which this taxpayer has been legally active.', verbose_name='active since'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='taxpayerprofile',
            name='issuing_email',
            field=models.TextField(blank=True, null=True, verbose_name='issuing email'),
        ),
        migrations.AddField(
            model_name='taxpayerprofile',
            name='sales_terms',
            field=models.CharField(default=None, help_text='The terms of the sale printed onto receipts by default (eg: single payment, checking account, etc).', max_length=48, verbose_name='sales terms'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='pointofsales',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='points_of_sales', to='afip.TaxPayer', verbose_name='owner'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='client_name',
            field=models.CharField(max_length=128, verbose_name='client name'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='issuing_email',
            field=models.CharField(max_length=128, null=True, verbose_name='issuing email'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='issuing_name',
            field=models.CharField(max_length=128, verbose_name='issuing name'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='pdf_file',
            field=models.FileField(blank=True, null=True, upload_to='', verbose_name='pdf file'),
        ),
    ]
